import json
from collections import UserDict
from collections.abc import Hashable
from typing import Callable, Self, Iterable, Mapping

from oldaplib.src.enums.action import Action
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.helpers.serializer import serializer


@serializer
class ObservableDict(UserDict):
    __on_change: Callable[[Self], None]
    _changeset: dict[Hashable, AttributeChange]

    def __init__(self,
                 obj: Iterable | Mapping | None = None, *,
                 on_change: Callable[[Self], None] | None = None,
                 obsdict: dict | None = None,
                 validate: bool = False,
                 **kwargs):
        """
        Initializes a new instance of the class with optional data and settings for
        notifications and validation.

        :param obj: An optional iterable or mapping containing initial data for the
            instance. Defaults to None.
        :type obj: Iterable | Mapping | None
        :param on_change: A callable function that will be invoked when a change is made
            to the instance. The function is passed the instance itself as an argument.
            Defaults to None.
        :type on_change: Callable[[Self], None] | None
        :param obsdict: This is used in conjunction with the method _as_dict to serialize
            the instance's data to JSON and back. It preserves the dataclasses also of the keys.
            *NOTE": It should never be used directly – it's reserved for the @serializer decorator.
        :type obsdict: dict | None
        :param validate: A boolean flag indicating whether to enforce validation checks
            on the data. Defaults to False.
        :type validate: bool
        :param kwargs: Additional keyword arguments forwarded to the superclass
            initializer.
        """
        self.__on_change = on_change
        self._changeset = {}
        if obj:
            super().__init__(obj, **kwargs)
        else:
            super().__init__(**kwargs)
        if obsdict:
            for item in obsdict:
                self[item['key']] = item['val']

    def __setitem__(self, key, value) -> None:
        if key in self.data:
            self._changeset[key] = AttributeChange(self.data[key], Action.MODIFY)
        else:
            self._changeset[key] = AttributeChange(None, Action.CREATE)
        if self.__on_change:
            self.__on_change(self.copy())
        super().__setitem__(key, value)

    def __delitem__(self, key) -> None:
        self._changeset[key] = AttributeChange(self.data[key], Action.DELETE)
        if self.__on_change:
            self.__on_change(self.copy())
        super().__delitem__(key)

    def __bool__(self) -> bool:
        return len(self) > 0

    def copy(self) -> Self:
        return ObservableDict(self.data.copy())

    def set_on_change(self, on_change: Callable[[Self], None]) -> None:
        self.__on_change = on_change

    def _as_dict(self):
        return {'obsdict': [{'key': key, 'val': val} for key, val in self.data.items()]}

    def clear_changeset(self) -> None:
        for item in self.data.values():
            if hasattr(item, 'clear_changeset'):
                item.clear_changeset()
        self._changeset = {}




from copy import deepcopy
from enum import Enum
from functools import partial
from typing import Callable, Self, Any

from oldaplib.src.enums.action import Action
from oldaplib.src.enums.attributeclass import AttributeClass
from oldaplib.src.enums.haspropertyattr import HasPropertyAttr
from oldaplib.src.enums.owlpropertytype import OwlPropertyType
from oldaplib.src.helpers.Notify import Notify
from oldaplib.src.helpers.irincname import IriOrNCName
from oldaplib.src.helpers.oldaperror import OldapErrorNotFound, OldapErrorInconsistency
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.project import Project
from oldaplib.src.propertyclass import PropertyClass, HasPropertyData
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_nonnegativeinteger import Xsd_nonNegativeInteger
from oldaplib.src.xsd.xsd_qname import Xsd_QName


@serializer
class PropType(Enum):
    INTERNAL = 1
    STANDALONE = 2
    EXTERNAL = 3


@serializer
class HasProperty(Model, Notify):
    """
    Represents a `HasProperty` model that integrates functionalities from both the `Model`
    and `Notify` classes. It provides extensive management for property attributes,
    notifications, and serialization.

    This class serves the purpose of managing `PropertyClass` objects or equivalent IRIs
    within the context of a given project. It offers capabilities such as notifying
    changes, serializing properties into dictionaries, and generating SHACL and OWL
    annotations for validation and reasoning purposes. The class also helps track
    updates to attributes or changes made to properties associated with a project.

    :ivar prop: The associated `PropertyClass` object or IRI of this property.
    :type prop: PropertyClass | Iri | None
    :ivar haspropdata: Represents combined data attributes (minCount, maxCount, order, group)
                      as a single `HasPropertyData` object.
    :type haspropdata: HasPropertyData
    """


    _prop: PropertyClass | Xsd_QName | None
    _project: Project | None
    _type: PropType | None

    def __init__(self, *,
                 con: IConnection,
                 project: Project | Iri | Xsd_NCName | str,
                 prop: PropertyClass | Xsd_QName | None = None,
                 notifier: Callable[[Xsd_QName], None] | None = None,
                 notify_data: Xsd_QName | None = None,
                 creator: Iri | None = None,  # DO NO USE! Only for jsonify!!
                 created: Xsd_dateTime | None = None,  # DO NO USE! Only for jsonify!!
                 contributor: Iri | None = None,  # DO NO USE! Only for jsonify!!
                 modified: Xsd_dateTime | None = None,  # DO NO USE! Only for jsonify!!
                 validate: bool = False,
                 _type: PropType | None = None,
                 **kwargs):
        """
        Initializes the class with the given parameters and configurations.

        This constructor sets up the internal properties and state, and handles specific
        initialization logic related to `PropertyClass`, notification handlers, and attributes.
        It also prepares the object to track changes and attributes defined in `HasPropertyAttr`.

        :param con: The connection interface used to interact with the related resources.
        :type con: IConnection
        :param project: The project instance associated with this object.
        :type project: Project
        :param prop: Optional; the property class or IRI associated with this object, or None.
        :param notifier: Optional; the callable function used for notification.
        :type notifier: Callable[[Iri], None] | None
        :param notify_data: Optional; additional data in the form of an IRI for notification.
        :type notify_data: Iri | None
        :param kwargs: Additional attributes to configure specific behaviors, passed as key-value pairs.

        :raises OldapErrorNotFound: If the given property class is not found.
        """
        Model.__init__(self, connection=con,
                       creator=creator,
                       created=created,
                       contributor=contributor,
                       modified=modified)
        Notify.__init__(self, notifier, notify_data)
        self._type = _type
        if isinstance(project, Project):
            self._project = project
        else:
            if not isinstance(project, (Iri, Xsd_NCName)):
                project = IriOrNCName(project, validate=validate)
            self._project = Project.read(self._con, project)
        if isinstance(prop, Xsd_QName):
            fixed_prop = Xsd_QName(str(prop).removesuffix("Shape"))
            try:
                self._prop = PropertyClass.read(self._con, self._project, fixed_prop)
                self._type = PropType.STANDALONE
            except OldapErrorNotFound as err:
                self._prop = fixed_prop
                self._type = PropType.EXTERNAL
        else:
            self._prop = prop
            if self._type is None:
                if prop.externalOntology:
                    self._type = PropType.EXTERNAL
                else:
                    self._type = PropType.INTERNAL
        self.set_attributes(kwargs, HasPropertyAttr)

        #
        # Check consistency for owl:FunctionalProperty. It must have a maxCount=1
        #
        if isinstance(self._prop, PropertyClass) and OwlPropertyType.FunctionalProperty in self._prop.type:
            if not self._attributes.get(HasPropertyAttr.MAX_COUNT):
                raise OldapErrorInconsistency(f'FunctionalProperty {self._prop.property_class_iri} must have maxCount=1')
            if self._attributes[HasPropertyAttr.MAX_COUNT] != 1:
                raise OldapErrorInconsistency(f'FunctionalProperty {self._prop.property_class_iri} must have maxCount=1')

        #
        # Check consistency for owl:InverseFunctionalProperty. It must have a cardinality of 1..1
        #
        if isinstance(self._prop, PropertyClass) and OwlPropertyType.InverseFunctionalProperty in self._prop.type:
            if not self._attributes.get(HasPropertyAttr.MIN_COUNT) or not self._attributes.get(HasPropertyAttr.MAX_COUNT):
                raise OldapErrorInconsistency(f'InverseFunctionalProperty {self._prop.property_class_iri} must have cardinality=1..1')
            if self._attributes[HasPropertyAttr.MIN_COUNT] != 1 or self._attributes[HasPropertyAttr.MAX_COUNT] != 1:
                raise OldapErrorInconsistency(f'InverseFunctionalProperty {self._prop.property_class_iri} must have cardinality=1..1')


        for attr in HasPropertyAttr:
            setattr(HasProperty, attr.value.fragment, property(
                partial(HasProperty._get_value, attr=attr),
                partial(HasProperty._set_value, attr=attr),
                partial(HasProperty._del_value, attr=attr)))
        self.update_notifier(notifier, notify_data)
        self._changeset = {}

    def update_notifier(self,
                        notifier: Callable[[AttributeClass | Xsd_QName], None] | None = None,
                        notify_data: HasPropertyAttr | Xsd_QName | None = None):
        self.set_notifier(notifier, notify_data)
        if isinstance(self._prop, PropertyClass):
            self._prop.set_notifier(self.notifier, self._prop.property_class_iri)
        for attr, value in self._attributes.items():
            if getattr(value, 'set_notifier', None) is not None:
                value.set_notifier(self.notifier, attr)

    def _as_dict(self):
        return {x.fragment: y for x, y in self._attributes.items()} | super()._as_dict() | {
            'project': self._project.projectShortName,
            '_type': self._type,
            'prop': self._prop
        }

    def __deepcopy__(self, memo: dict[Any, Any]) -> Self:
        if id(self) in memo:
            return memo[id(self)]
        cls = self.__class__
        instance = cls.__new__(cls)
        memo[id(self)] = instance
        Model.__init__(instance,
                       connection=deepcopy(self._con, memo),
                       creator=deepcopy(self._creator, memo),
                       created=deepcopy(self._created, memo),
                       contributor=deepcopy(self._contributor, memo),
                       modified=deepcopy(self._modified, memo))
        Notify.__init__(instance,
                        notifier=self._notifier,
                        data=deepcopy(self._notify_data, memo))
        # Copy internals of Model:
        instance._attributes = deepcopy(self._attributes, memo)
        instance._changset = deepcopy(self._changeset, memo)
        instance._prop = deepcopy(self._prop)  # no deepcopy here (????)
        return instance

    def __str__(self) -> str:
        iri = self._prop.property_class_iri if isinstance(self._prop, PropertyClass) else self._prop
        s = f'HasProperty: {iri}\n'
        s += Model.__str__(self)
        return s

    @property
    def type(self) -> PropType:
        return self._type

    @property
    def prop(self) -> PropertyClass | Xsd_QName | None:
        return self._prop

    @prop.setter
    def prop(self, prop: PropertyClass | Xsd_QName) -> None:
        self._prop = prop

    @property
    def haspropdata(self) -> HasPropertyData:
        return HasPropertyData(minCount=self._attributes.get(HasPropertyAttr.MIN_COUNT, None),
                               maxCount=self._attributes.get(HasPropertyAttr.MAX_COUNT, None),
                               order=self._attributes.get(HasPropertyAttr.ORDER, None),
                               group=self._attributes.get(HasPropertyAttr.GROUP, None))

    def clear_changeset(self) -> None:
        if hasattr(self._prop, 'clear_changeset'):
            self._prop.clear_changeset()
        super().clear_changeset()

    def notifier(self, attr: HasPropertyAttr | Iri | Xsd_QName) -> None:
        #if attr == HasPropertyAttr.PROP:
        #    return
        if isinstance(attr, HasPropertyAttr):
            self._changeset[attr] = AttributeChange(self._attributes[attr], Action.MODIFY)
        elif isinstance(attr, Xsd_QName):
            self._changeset[attr] = AttributeChange(None, Action.MODIFY)
        self.notify()

    def create_shacl(self, indent: int = 0, indent_inc: int = 4) -> str:
        blank = ''
        sparql = ''
        if self._attributes.get(HasPropertyAttr.MIN_COUNT, None) is not None:
            sparql += f' ;\n{blank:{indent * indent_inc}}sh:minCount {self._attributes[HasPropertyAttr.MIN_COUNT].toRdf}'
        if self._attributes.get(HasPropertyAttr.MAX_COUNT, None) is not None:
            sparql += f' ;\n{blank:{indent * indent_inc}}sh:maxCount {self._attributes[HasPropertyAttr.MAX_COUNT].toRdf}'
        if self._attributes.get(HasPropertyAttr.ORDER, None) is not None:
            sparql += f' ;\n{blank:{indent * indent_inc}}sh:order {self._attributes[HasPropertyAttr.ORDER].toRdf}'
        if self._attributes.get(HasPropertyAttr.GROUP, None) is not None:
            sparql += f' ;\n{blank:{indent * indent_inc}}sh:group {self._attributes[HasPropertyAttr.GROUP].toRdf}'
        return sparql

    def create_owl(self, indent: int = 0, indent_inc: int = 4) -> str:
        blank = ''
        sparql = ''
        min_count = Xsd_nonNegativeInteger(int(self._attributes[HasPropertyAttr.MIN_COUNT])) if self._attributes.get(HasPropertyAttr.MIN_COUNT) else None
        max_count = Xsd_nonNegativeInteger(int(self._attributes[HasPropertyAttr.MAX_COUNT])) if self._attributes.get(HasPropertyAttr.MAX_COUNT) else None

        if min_count and max_count and min_count == max_count:
            tmp = Xsd_nonNegativeInteger(min_count)
            sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}owl:qualifiedCardinality {tmp.toRdf}'
        else:
            if min_count:
                tmp = Xsd_nonNegativeInteger(min_count)
                sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}owl:minQualifiedCardinality {tmp.toRdf}'
            if max_count:
                tmp = Xsd_nonNegativeInteger(max_count)
                sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}owl:maxQualifiedCardinality {tmp.toRdf}'
        return sparql

    def update_shacl(self,
                     graph: Xsd_NCName,
                     resclass_iri: Xsd_QName,
                     propclass_iri: Xsd_QName,
                     indent: int = 0, indent_inc: int = 4) -> str:
        blank = ''
        sparql_list = []
        for attr, change in self._changeset.items():
            if isinstance(attr, Xsd_QName):  # if it's an IRI, the attached PropertyClass has changed. We don't process this here
                continue
            sparql = f'WITH {graph}:shacl\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'

            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} {self._attributes[attr].toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'

            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}{resclass_iri}Shape sh:property ?prop .\n'
            if isinstance(self._prop, Xsd_QName) or self._prop.internal is None:
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop sh:node {propclass_iri}Shape.\n'
            else:
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop sh:path {propclass_iri} .\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} {change.old_value.toRdf} .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)
        sparql = " ;\n".join(sparql_list)
        return sparql

    def update_owl(self,
                     graph: Xsd_NCName,
                     resclass_iri: Xsd_QName,
                     propclass_iri: Xsd_QName,
                     indent: int = 0, indent_inc: int = 4) -> str:
        blank = ''
        if HasPropertyAttr.MIN_COUNT in self._changeset or HasPropertyAttr.MAX_COUNT in self._changeset:
            sparql = f'WITH {graph}:onto\n'
            sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:qualifiedCardinality ?val_qualified .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:minQualifiedCardinality ?val_min .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:maxQualifiedCardinality ?val .\n'
            sparql += f'{blank:{indent * indent_inc}}}}\n'

            #
            # Convert to xsd:nonNegativeInteger or None
            #
            min_count = Xsd_nonNegativeInteger(int(self._attributes[HasPropertyAttr.MIN_COUNT])) if self._attributes.get(HasPropertyAttr.MIN_COUNT) else None
            max_count = Xsd_nonNegativeInteger(int(self._attributes[HasPropertyAttr.MAX_COUNT])) if self._attributes.get(HasPropertyAttr.MAX_COUNT) else None

            if min_count or max_count:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                if min_count and max_count and min_count == max_count:
                    tmp = Xsd_nonNegativeInteger(min_count)
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:qualifiedCardinality {tmp.toRdf} .\n'
                else:
                    if min_count:
                        tmp = Xsd_nonNegativeInteger(min_count)
                        sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:minQualifiedCardinality {tmp.toRdf} .\n'
                    if max_count:
                        tmp = Xsd_nonNegativeInteger(max_count)
                        sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:maxQualifiedCardinality {tmp.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'

            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}{resclass_iri} rdfs:subClassOf ?prop .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:onProperty {propclass_iri} .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}OPTIONAL {{ ?prop owl:qualifiedCardinality ?val_qualified . }}\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}OPTIONAL {{ ?prop owl:minQualifiedCardinality ?val_min . }}\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}OPTIONAL {{ ?prop owl:maxQualifiedCardinality ?val_max . }}\n'
            sparql += f'{blank:{indent * indent_inc}}}}'

            return sparql
        else:
            return ''



import json
import os
from time import sleep

import bcrypt
import jwt
import requests

from jwt import InvalidTokenError
from typing import Dict, Optional, Any
from datetime import datetime, timedelta
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore
from pathlib import Path

from requests.auth import HTTPBasicAuth

from oldaplib.src.oldaplogging import get_logger
from oldaplib.src.version import __version__

from oldaplib.src.cachesingleton import CacheSingleton, CacheSingletonRedis
from oldaplib.src.enums.adminpermissions import AdminPermission
from oldaplib.src.userdataclass import UserData
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorNoPermission
from oldaplib.src.helpers.context import Context, DEFAULT_CONTEXT
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.iconnection import IConnection
from oldaplib.src.enums.sparql_result_format import SparqlResultFormat
from oldaplib.src.xsd.xsd_string import Xsd_string

#
# For bootstrapping the whole tripel store, the following SPARQL has to be executed within the GraphDB
# SPARQL-console:
#
"""
"""
#
# Then, executing the __main__ of the file "connection.py" will initialize the triple store with all the data
# needed to run the tests
#
jwt_format = {
    "userId": "https://orcid.org/0000-0003-1681-4036",
    "exp": "2023-11-04T12:00:00+00:00",
    "iat": datetime.now().astimezone().isoformat(),
    "iss": "http://oldap.org"
}
token = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VySWQiOiIkMmIkMTIkaldDSloucWRYRTlNU0NQZFVjMHk0Ljlzd1dZSmNnTFpuMGVQdFJUdS83VThxSC9PWFhrQjIiLCJleHAiOiIyMDI0LTExLTA0VDEyOjAwOjAwKzAwOjAwIiwiaWF0IjoiMjAyNC0wMS0xOVQyMzo0MTozMS45NTI5MTkiLCJpc3MiOiJodHRwOi8vb2xkYXAub3JnIn0.Vsc2qamfyeTW6Xz5l2Wca-mFnA5PcLuOoWPVEo__4Z4"


class Connection(IConnection):
    """
    Class that implements the connection to an external triple store for oldap.

    The connection to a SPARQL endpoint requires the following information:

    * _server_: URL of the server (inlcuding port number)
    * _repo_: Name of the repository to connect to
    * _context_name_: Name of the context (see ~helper.Context)

    The class implements the following methods:

    * Getter-methods:
        - _server_: returns the server string
        - _repo_: returns the repository name
        - _context_name_: returns the context name
    * Setter methods:
        - Any setting of the _server_, _repo_ or _context_name_ - variables raises an OldapError-exception
    * Further methods
        - _Constructor(server,repo,contextname)_: requires _server_ and _repo_string, _context_name defaults to "DEFAULT"
        - _clear_graph_(graph_name: QName)_: Deletes the given graph (must be given as QName)
        - _clear_repo()_ Deletes all data in the repository given by the Connection instance
        - _upload_turtle(filename: str, graphname:str)_: Loads the data in the given file (must be turtle or trig
          format) into the given repo and graph. If graphname is not given, the data will either be loaded into the
          default graph of the repository or into the graph given in the trig file.
          _Note_: The method returns before the triple store has digested all the data! It may not immediately
          available after this method returns!
        - _query(query: str, format: SparqlResultFormat)_: Sends a SPARQL query to the triple store and returns the
          result in the given format. If no format is given, JSON will be used as default format.
        - _update_query(query: str)_: Send a SPARQL update query to the SPARQL endpoint. The method return either
          {'status': 'OK'} or {'status': 'ERROR', 'message': 'error-text'}
        - _rdflib_query(query: str, bindings: Optional[Mapping[str, Identifier]])_: Send a SPAQRL query using rdflib
          to the SPARQL endpoint. The variable _bindings_ allows to set query parameters to given values.

    """
    _server: str
    _repo: str
    _userId: str
    _dbuser: str
    _dbpassword: str
    _userdata: Optional[UserData]
    _token: Optional[str]
    _context_name: str = DEFAULT_CONTEXT
    _store: SPARQLUpdateStore
    _query_url: str
    _update_url: str
    _transaction_url: Optional[str]
    __jwtkey: str
    _switcher = {
        SparqlResultFormat.XML: lambda a: a.text,
        SparqlResultFormat.JSON: lambda a: a.json(),
        SparqlResultFormat.TURTLE: lambda a: a.text,
        SparqlResultFormat.N3: lambda a: a.text,
        SparqlResultFormat.NQUADS: lambda a: a.text,
        SparqlResultFormat.JSONLD: lambda a: a.json(),
        SparqlResultFormat.TRIX: lambda a: a.text,
        SparqlResultFormat.TRIG: lambda a: a.text,
        SparqlResultFormat.TEXT: lambda a: a.text
    }

    def __init__(self, *,
                 server: Optional[str] = None,
                 repo: Optional[str] = None,
                 userId: Optional[str | Xsd_NCName] = None,
                 credentials: Optional[str | Xsd_string] = None,
                 token: Optional[str] = None,
                 dbuser: Optional[str] = None,
                 dbpassword: Optional[str] = None,
                 context_name: Optional[str] = DEFAULT_CONTEXT) -> None:
        """
        Constructor that establishes the connection parameters.

        :param server: URL of the server (including port information if necessary)
        :type server: str
        :param repo: Name of the triple store repository on the server
        :type repo: str
        :param userId: User identifier for authentication. If not specified, a default "unknown" value
                       is used. Can optionally be validated against `Xsd_NCName`.
        :type userId: Optional[str | Xsd_NCName]
        :param credentials: User credentials for authentication. If not specified, defaults to None.
        :type credentials: Optional[str]
        :param token: JWT token string for authentication. If provided, it bypasses userId and credentials.
        :type token: Optional[str]
        :param context_name: A name of the Context to be used (see ~Context). If no such context exists,
                             a new context with this name is created.
        :type context_name: Optional[str]
        :raises OldapError: Raised when invalid credentials or token are provided, or if there is
                            an issue during the authentication process. Also raised on login failure
                            in specific scenarios.
        """
        super().__init__(context_name=context_name)
        self.__jwtkey = os.getenv("OLDAP_JWT_SECRET", "You have to change this!!! +D&RWG+")
        self._server = server or os.getenv("OLDAP_TS_SERVER", "http://localhost:7200")
        self._repo = repo or os.getenv("OLDAP_TS_REPO", "oldap")
        self._dbuser = dbuser or os.getenv("OLDAP_TS_USER", "")
        self._dbpassword = dbpassword or os.getenv("OLDAP_TS_PASSWORD", "")
        self._query_url = f'{self._server}/repositories/{self._repo}'
        self._update_url = f'{self._server}/repositories/{self._repo}/statements'
        self._store = SPARQLUpdateStore(self._query_url, self._update_url)

        logger = get_logger()

        context = Context(name=context_name)
        if token is not None:
            try:
                payload = jwt.decode(jwt=token, key=self.__jwtkey, algorithms="HS256")
            except InvalidTokenError:
                logger.error("Connection with invalid token")
                raise OldapError("Wrong credentials")
            self._userdata = json.loads(payload['userdata'], object_hook=serializer.decoder_hook)
            self._token = token
            return
        if userId is None and credentials is None:
            userId = Xsd_NCName("unknown", validate=False)
        if not isinstance(userId, Xsd_NCName):
            userId = Xsd_NCName(userId)
        if userId is None:
            logger.error("Connection with wrong credentials")
            raise OldapError("Wrong credentials")
        sparql = UserData.sparql_query(context=context, userId=userId)
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Accept": "application/x-sparqlstar-results+json, application/sparql-results+json;q=0.9, */*;q=0.8",
        }
        data = {
            'query': sparql,
        }
        #
        # if we have protected the triplestore by a user/password, add it to the request
        #
        auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
        res = requests.post(url=self._query_url, headers=headers, data=data, auth=auth)
        if res.status_code == 200:
            jsonobj = res.json()
        else:
            logger.error(f"Could not connect to triplestore: {res.text}")
            raise OldapError(res.status_code, res.text)
        res = QueryProcessor(context=context, query_result=jsonobj)

        self._userdata = UserData.from_query(res)
        if not self._userdata.isActive:
            logger.error("Connection with wrong credentials")
            raise OldapError("Wrong credentials")  # On purpose, we are not providing too much information why the login failed
        if userId != "unknown":
            hashed = str(self._userdata.credentials).encode('utf-8')
            if not bcrypt.checkpw(credentials.encode('utf-8'), hashed):
                logger.error("Connection with wrong credentials")
                raise OldapError("Wrong credentials")  # On purpose, we are not providing too much information why the login failed

        expiration = datetime.now().astimezone() + timedelta(days=1)
        payload = {
            "userdata": json.dumps(self._userdata, default=serializer.encoder_default),
            "exp": expiration.timestamp(),
            "iat": int(datetime.now().astimezone().timestamp()),
            "iss": "http://oldap.org"
        }
        self._token = jwt.encode(
            payload=payload,
            key=self.__jwtkey,
            algorithm="HS256")
        #
        # Get projects and add to Context
        #
        sparql = context.sparql_context
        sparql += """
        SELECT ?sname ?ns
        FROM oldap:onto
        FROM shared:onto
        FROM NAMED oldap:admin
        WHERE {
            GRAPH oldap:admin {
                ?proj a oldap:Project .
                ?proj oldap:projectShortName ?sname .
                ?proj oldap:namespaceIri ?ns .
            }
        }
        """
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Accept": "application/x-sparqlstar-results+json, application/sparql-results+json;q=0.9, */*;q=0.8",
        }
        data = {
            'query': sparql,
        }
        res = requests.post(url=self._query_url, headers=headers, data=data, auth=auth)
        if res.status_code == 200:
            jsonobj = res.json()
        else:
            logger.error(f"Could not connect to triplestore: {res.text}")
            raise OldapError(res.status_code, res.text)
        res = QueryProcessor(context=context, query_result=jsonobj)
        for r in res:
            context[r['sname']] = r['ns']
        logger.info(f'Connection established. User "{str(self._userdata.userId)}".')

    @staticmethod
    def version(self) -> str:
        return __version__

    @property
    def jwtkey(self) -> str:
        """Getter for the JWT token"""
        return self.__jwtkey

    @property
    def server(self) -> str:
        """Getter for server string"""
        return self._server

    @property
    def repo(self) -> str:
        """Getter for repository name"""
        return self._repo

    def clear_graph(self, graph_iri: Xsd_QName) -> None:
        """
        Clears (deletes) the given RDF graph from the system.

        This method uses the SPARQL Update protocol to clear the specified graph.
        Permission checks are performed to ensure that the actor has sufficient
        privileges. If the user does not have the required permissions or
        the graph clearing operation fails, appropriate exceptions are raised.

        :param graph_iri: RDF graph name as QName. The prefix must be defined
                          in the context.
        :type graph_iri: Xsd_QName
        :return: None
        :rtype: None
        :raises OldapErrorNoPermission: If the user lacks the required
                                        permission to clear the graph.
        :raises OldapError: If the SPARQL update operation fails.
        """
        logger = get_logger()
        if not self._userdata:
            logger.error("Connection with no permission to clear graph.")
            raise OldapErrorNoPermission("No permission")
        actor = self._userdata
        sysperms = actor.inProject.get(Xsd_QName('oldap:SystemProject'))
        is_root: bool = False
        if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
            is_root = True
        if not is_root:
            raise OldapErrorNoPermission("No permission")

        context = Context(name=self._context_name)
        headers = {
            "Content-Type": "application/sparql-update",
            "Accept": "application/json, text/plain, */*",
        }
        data = f"CLEAR GRAPH <{context.qname2iri(graph_iri)}>"
        auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
        req = requests.post(self._update_url,
                            headers=headers,
                            data=data,
                            auth=auth)
        if not req.ok:
            logger.error(f'Clearing of graph "{graph_iri}" failed: {req.text}')
            raise OldapError(req.text)
        logger.info(f'Graph "{graph_iri}" cleared.')

    def clear_repo(self) -> None:
        """
        Deletes the complete repository. This operation is destructive and should
        be used with extreme caution. Removes all data from the repository and executes
        a "CLEAR ALL" update operation.

        :return: None
        """
        # if not self._userdata:
        #     raise OldapErrorNoPermission("No permission")
        # actor = self._userdata
        # sysperms = actor.inProject.get(QName('oldap:SystemProject'))
        # is_root: bool = False
        # if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
        #     is_root = True
        # if not is_root:
        #     raise OldapErrorNoPermission("No permission")
        headers = {
            "Accept": "application/json, text/plain, */*",
        }
        data = {"update": "CLEAR ALL"}
        auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
        req = requests.post(self._update_url,
                            headers=headers,
                            data=data,
                            auth=auth)
        if not req.ok:
            raise OldapError(req.text)

    def upload_turtle(self, filename: str, graphname: Optional[str] = None) -> None:
        """
        Uploads a turtle or trig file to the specified repository. This function sends the file to the triplestore for
        import and does not wait for the completion of the import process. The process itself may continue for some time
        after the command is issued.

        :param filename: Name of the file to be uploaded.
        :type filename: str
        :param graphname: Optional; the name of the RDF-graph into which the data is to be imported.
        :type graphname: str or None
        :return: None
        :rtype: None
        :raises OldapError: Raised when there are issues with the repository or during the HTTP request.
        """
        # if not self._userdata:
        #     raise OldapErrorNoPermission("No permission")
        # actor = self._userdata
        # sysperms = actor.inProject.get(QName('oldap:SystemProject'))
        # is_root: bool = False
        # if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
        #     is_root = True
        # if not is_root:
        #     raise OldapErrorNoPermission("No permission")

        logger = get_logger()
        with open(filename, encoding="utf-8") as f:
            content = f.read()
            ext = Path(filename).suffix
            mime = ""
            if ext == ".ttl":
                mime = "text/turtle"
            elif ext == ".trig":
                mime = "application/x-trig"

            ct = datetime.now().astimezone()
            ts = ct.timestamp()
            data = {
                "name": f'Data from "{filename}"',
                "status": None,
                "message": "",
                "context": graphname,
                "replaceGraphs": [],
                "baseURI": None,
                "forceSerial": False,
                "type": "text",
                "format": mime,
                "data": content,
                "timestamp": ts,
                "parserSettings": {
                    "preserveBNodeIds": False,
                    "failOnUnknownDataTypes": False,
                    "verifyDataTypeValues": False,
                    "normalizeDataTypeValues": False,
                    "failOnUnknownLanguageTags": False,
                    "verifyLanguageTags": True,
                    "normalizeLanguageTags": True,
                    "stopOnError": True
                }
            }
            jsondata = json.dumps(data)
            headers = {
                "Accept": "application/json, text/plain, */*",
                "Content-Type": "application/json; charset=utf-8"
            }
            url = f"{self._server}/rest/repositories/{self._repo}/import/upload/text"
            auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
            req = requests.post(url,
                                headers=headers,
                                data=jsondata,
                                auth=auth)
        if not req.ok:
            logger.error(f'Upload of file "{filename}" failed: {req.text}')
            raise OldapError(req.text)
        logger.info(f'File "{filename}" uploaded.')

    def query(self, query: str, format: SparqlResultFormat = SparqlResultFormat.JSON) -> Any:
        """
        Send a SPARQL-query and return the result. The result may be nested dict (in case of JSON) or a text.

        :param query: SPARQL query as string
        :type query: str
        :param format: The format desired (see ~SparqlResultFormat)
        :type format: SparqlResultFormat
        :return: Query results or an error message (as text)
        :rtype: Any
        :raises OldapError: Raised if not logged in or if there is an issue with the query execution.
        """
        logger = get_logger()
        if not self._userdata:
            logger.error("Not a valid user session.")
            raise OldapError("No login")
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Accept": format.value,
        }
        data = {
            'query': query,
        }
        auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
        res = requests.post(url=self._query_url,
                            headers=headers,
                            data=data,
                            auth=auth)
        if res.status_code == 200:
            return Connection._switcher[format](res)
        else:
            logger.error(f"SPARQL query failed: {res.text}")
            raise OldapError(res.text)

    def update_query(self, query: str) -> Dict[str,str]:
        """
        Sends an SPARQL UPDATE query to the triple store.

        This method constructs, performs, and handles the response of an SPARQL UPDATE query
        sent to an RDF triple store. If no user session is active, an error will be raised.

        :param query: The SPARQL UPDATE query as a string to be executed on the triple store.
        :type query: str
        :return: A dictionary containing the response information from the triple store.
        :rtype: Dict[str, str
        :raises OldapError: If user authentication is missing or the SPARQL UPDATE execution fails.
        """
        logger = get_logger()
        if not self._userdata:
            logger.error("Not a valid user session.")
            raise OldapError("No login")
        headers = {
            "Accept": "*/*"
        }
        url = f"{self._server}/repositories/{self._repo}/statements"
        auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
        res = requests.post(url, data={"update": query}, headers=headers, auth=auth)
        if not res.ok:
            logger.error(f"SPARQL update query failed: {res.text}")
            raise OldapError(f'Update query failed. Reason: "{res.text}"')

    def transaction_start(self) -> None:
        """
        Initiates a new transaction for the current repository on the server.

        This method starts a transaction by sending a POST request to the server's
        transaction endpoint. If the `location` header is missing in the response,
        it indicates that the transaction initiation failed.

        :raises OldapError: If the user is not logged in or if the transaction cannot
                            be started on the server.
        :return: None
        """
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Accept": "*/*"
        }
        url = f"{self._server}/repositories/{self._repo}/transactions"
        auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
        res = requests.post(url, headers=headers, auth=auth)
        if res.headers.get('location') is None:
            raise OldapError('GraphDB start of transaction failed')
        self._transaction_url = res.headers['location']

    def transaction_query(self, query: str, result_format: SparqlResultFormat = SparqlResultFormat.JSON) -> Any:
        """
        Executes a SPARQL query against the currently ongoing transaction in GraphDB.

        This method sends a SPARQL query using HTTP POST to the active transaction URL
        and handles the response based on the provided result format. It ensures that
        the user is authenticated and that a transaction is already started. In case
        of errors, appropriate exceptions are raised.

        :param query: The SPARQL query as a string to execute.
        :type query: str
        :param result_format: The expected format of the SPARQL query result. Defaults
                              to JSON.
        :type result_format: SparqlResultFormat
        :return: The query result in the specified format.
        :rtype: Any
        :raises OldapError: If no user is logged in, no transaction is started, or the
                            query execution fails.
        """
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Accept": result_format.value
        }
        if self._transaction_url is None:
            raise OldapError("No GraphDB transaction started")
        auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
        res = requests.post(self._transaction_url,
                            data={'action': 'QUERY', 'query': query},
                            headers=headers,
                            auth=auth)
        if not res.ok:
            raise OldapError(f'GraphDB Transaction query failed. Reason: "{res.text}"')
        return Connection._switcher[result_format](res)

    def transaction_update(self, query: str) -> None:
        """
        Updates the current GraphDB transaction with a specified SPARQL update query.

        This method sends a POST request to the specified transaction URL with the
        provided SPARQL `query`. It throws an error if the user is not logged in or
        if no GraphDB transaction URL is available. If the request fails, an exception
        is raised containing the reason for the failure.

        :param query: The SPARQL update query to execute as part of the current GraphDB
                      transaction.
        :type query: str
        :raises OldapError: If the user is not logged in or no transaction URL is defined.
        :raises OldapError: If the GraphDB transaction update fails.
        :return: None
        """
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Accept": "*/*"
        }
        if self._transaction_url is None:
            raise OldapError("No GraphDB transaction started")
        auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
        res = requests.post(self._transaction_url,
                            data={'action': 'UPDATE', 'update': query},
                            headers=headers,
                            auth=auth)
        if not res.ok:
            raise OldapError(f'GraphDB Transaction update failed. Reason: "{res.text}"')

    def transaction_commit(self) -> None:
        """
        Commits the current transaction for the GraphDB session. This method ensures that
        the transaction completes and is finalized successfully. In the case of an error
        in committing the transaction, it raises an exception to notify the caller.

        :raises OldapError: If the user is not logged in, no transaction is started,
            or the commit operation fails.
        :rtype: None
        :return: None
        """
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Accept": "*/*"
        }
        if self._transaction_url is None:
            raise OldapError("No GraphDB transaction started")
        auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
        res = requests.put(f'{self._transaction_url}?action=COMMIT', headers=headers, auth=auth)
        if not res.ok:
            raise OldapError(f'GraphDB transaction commit failed. Reason: "{res.text}"')
        self._transaction_url = None

    def transaction_abort(self) -> None:
        """
        Aborts an ongoing GraphDB transaction if it exists. This method ensures that
        a started transaction is properly terminated to prevent stale or lingering
        transactions.

        :raises OldapError: If the user is not logged in or there is no transaction
            to abort, or if the abort request to the server fails.
        :return: None
        """
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Accept": "*/*"
        }
        if self._transaction_url is None:
            raise OldapError("No GraphDB transaction started")
        auth = HTTPBasicAuth(self._dbuser, self._dbpassword) if self._dbuser and self._dbpassword else None
        res = requests.delete(self._transaction_url, headers=headers, auth=auth)
        if not res.ok:
            raise OldapError(f'GraphDB transaction abort failed. Reason: "{res.text}"')
        self._transaction_url = None

    def in_transaction(self) -> bool:
        """
        Determines if the current instance is in a transaction.

        This method checks if there is a transaction URL indicating that a transaction
        is currently active. If `_transaction_url` is not None, it means a transaction
        is in progress. Otherwise, no transaction is active.

        :return: True if in a transaction, False otherwise
        :rtype: bool
        """
        return self._transaction_url is not None


if __name__ == "__main__":
    con = Connection(server='http://localhost:7200',
                     userId="rosenth",
                     credentials="RioGrande",
                     repo="oldap",
                     context_name="DEFAULT")
    cache = CacheSingletonRedis()
    cache.clear()
    con.clear_repo()
    con.upload_turtle("../ontologies/oldap.trig")
    con.upload_turtle("../ontologies/shared.trig")
    con.upload_turtle("../ontologies/admin.trig")

    # Eingabeaufforderung für das Laden der Testinstanzen
    load_test_instances = input("Testinstanzen laden [Y/N] ?(Y): ").strip().lower()
    if not load_test_instances or load_test_instances in ['y', 'yes', 'ja']:
        print("Lade Testinstanzen...")
        con.upload_turtle("../ontologies/admin-testing.trig")
        print("Testinstanzen erfolgreich geladen.")
    else:
        print("Testinstanzen übersprungen.")



import re
from enum import Flag, auto
from functools import partial
from typing import Type, Any, Self, cast

from oldaplib.src.datamodel import DataModel
from oldaplib.src.enums.action import Action
from oldaplib.src.enums.datapermissions import DataPermission
from oldaplib.src.enums.haspropertyattr import HasPropertyAttr
from oldaplib.src.enums.adminpermissions import AdminPermission
from oldaplib.src.enums.propertyclassattr import PropClassAttr
from oldaplib.src.enums.xsd_datatypes import XsdDatatypes
from oldaplib.src.hasproperty import HasProperty
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.convert2datatype import convert2datatype
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.observable_set import ObservableSet
from oldaplib.src.helpers.oldaperror import OldapErrorNotFound, OldapErrorValue, OldapErrorInconsistency, \
    OldapErrorNoPermission, OldapError, OldapErrorUpdateFailed, OldapErrorInUse, OldapErrorAlreadyExists
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.iconnection import IConnection
from oldaplib.src.project import Project
from oldaplib.src.propertyclass import PropertyClass
from oldaplib.src.resourceclass import ResourceClass
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.xsd.xsd_datetimestamp import Xsd_dateTimeStamp
from oldaplib.src.xsd.xsd_integer import Xsd_integer
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_qname import Xsd_QName

ValueType = LangString | ObservableSet | Xsd

class SortBy(Flag):
    PROPVAL = auto()
    CREATED = auto()
    LASTMOD = auto()


#@strict
class ResourceInstance:
    """
    Represents an instance of a resource in a system based on certain defined ontology.

    Provides functionality to initialize a resource, set its values, validate those values
    against defined properties, and handle changesets related to the resource instance.
    It integrates the resource's superclass properties into the instance and ensures the
    conformance of its data to defined constraints such as minimum/maximum counts, lengths,
    and allowable values or patterns.

    Basically it's used be the ResourceInstanceFactory to create new instances of a resource.
    The basic usage is as follows:

    ```python
        project = Project.read(con=self._connection, projectIri_SName='test')
        factory = ResourceInstanceFactory(con=self._connection, project=project)
        Book = factory.createObjectInstance('Book')
    ```

    **IMPORTANT NOTE: This class is not intended to be used directly, but rather through the
    `ResourceInstanceFactory` class!**

    In a first step the factory is instantiated with the project. Then using the `createObjectInstance` method
    creates a new Python class which will represent instances of the given resource class.

    :ivar iri: The unique identifier of the resource instance.
    :type iri: Iri
    :ivar values: A mapping of property IRIs to their corresponding values, which could
                  include `LangString` objects or `ObservableSet` of values.
    :type values: dict[Iri, LangString | ObservableSet]
    :ivar graph: The name of the graph attribute in which the instance will reside,
                 represented by a short name of the project.
    :type graph: Xsd_NCName
    :ivar changeset: A mapping to track all attributes changed within the instance,
                     where the key is the property IRI and the value defines the
                     change details.
    :type changeset: dict[Iri, AttributeChange]
    """
    _iri: Iri
    _values: dict[Xsd_QName, LangString | ObservableSet]
    _graph: Xsd_NCName
    _changeset: dict[Iri, AttributeChange]

    __slots__ = ['_iri', '_values', '_graph', '_changeset', '_superclass_objs',
                 '_con', 'project', 'name', 'factory', 'properties', 'superclass']


    def __init__(self, *,
                 iri: Iri | None = None,
                 **kwargs):
        """
        Initializes an instance of the resource class, setting the optional IRI, processing
        superclasses, and validating property constraints. This initializer interprets
        the properties and superclasses associated with the instance and ensures
        conformance to rules such as `MIN_COUNT` and `MAX_COUNT` for properties.

        :param iri: An optional parameter specifying the IRI of the instance.
        :type iri: Iri | None
        :param kwargs: A dictionary of additional property values with property IRIs
            or fragments as keys. In case of unique fragments, the dictionary may be pass
            as named method arguments
        :type kwargs: dict or named method arguments
        """
        if iri and isinstance(iri, str):
            iri = Iri(Xsd_QName(self.project.projectShortName, iri))
        self._iri = Iri(iri, validate=True) if iri else Iri()
        self._values: dict[Xsd_QName, ValueType] = {}
        self._graph = self.project.projectShortName
        self._superclass_objs = {}
        self._changeset = {}

        def set_values(propclass: dict[Xsd_QName, HasProperty]):
            for prop_iri, hasprop in propclass.items():
                if kwargs.get(str(prop_iri)) or kwargs.get(prop_iri.fragment):
                    value = kwargs[str(prop_iri)] if kwargs.get(str(prop_iri)) else kwargs[prop_iri.fragment]
                    if isinstance(value, (list, tuple, set, LangString)):  # we may have multiple values...
                        if hasprop.prop.datatype == XsdDatatypes.langString:
                            self._values[prop_iri] = LangString(value,
                                                                notifier=self.notifier, notify_data=prop_iri)
                        else:
                            self._values[prop_iri] = ObservableSet({convert2datatype(x, hasprop.prop.datatype) for x in value},
                                                                   notifier=self.notifier, notify_data=prop_iri)
                    else:
                        try:
                            self._values[prop_iri] = ObservableSet({convert2datatype(value, hasprop.prop.datatype)})
                        except TypeError as err:
                            self._values[prop_iri] = convert2datatype(value, hasprop.prop.datatype)

            for prop_iri, hasprop in propclass.items():
                #
                # Validate
                #
                if hasprop.get(HasPropertyAttr.MIN_COUNT):  # testing for MIN_COUNT conformance
                    if hasprop[HasPropertyAttr.MIN_COUNT] > 0 and not self._values.get(prop_iri):
                        raise OldapErrorValue(f'{self.name}: Property {prop_iri} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} is missing')
                    elif isinstance(self._values[prop_iri], ObservableSet) and len(self._values[prop_iri]) < 1:
                        raise OldapErrorValue(f'{self.name}: Property {prop_iri} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} is missing')
                if hasprop.get(HasPropertyAttr.MAX_COUNT):  # testing for MAX_COUNT conformance
                    if isinstance(self._values.get(prop_iri), ObservableSet) and len(self._values[prop_iri]) > hasprop[HasPropertyAttr.MAX_COUNT]:
                        raise OldapErrorValue(f'{self.name}: Property {prop_iri} with MAX_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} has to many values (n={len(self._values[prop_iri])})')
                if self._values.get(prop_iri):
                    if isinstance(self._values[prop_iri], LangString):
                        self.validate_value(self._values[prop_iri], hasprop.prop)
                    else:
                        if isinstance(self._values[prop_iri], ObservableSet):
                            for val in self._values[prop_iri]:
                                self.validate_value(val, hasprop.prop)
                        else:
                            self.validate_value(self._values[prop_iri], hasprop.prop)

        def process_superclasses(superclass: dict[Xsd_QName, ResourceClass]):
            for sc_iri, sc in superclass.items():
                if not sc:
                    continue
                if sc.superclass:
                    process_superclasses(sc.superclass)
                if sc.owl_class_iri == Xsd_QName("oldap:Thing", validate=False):
                    timestamp = Xsd_dateTimeStamp()
                    if not self._values.get(Xsd_QName('oldap:createdBy', validate=False)):
                        self._values[Xsd_QName('oldap:createdBy', validate=False)] = ObservableSet({self._con.userIri})
                    if not self._values.get(Xsd_QName('oldap:creationDate', validate=False)):
                        self._values[Xsd_QName('oldap:creationDate', validate=False)] = ObservableSet({timestamp})
                    if not self._values.get(Xsd_QName('oldap:lastModifiedBy', validate=False)):
                        self._values[Xsd_QName('oldap:lastModifiedBy', validate=False)] = ObservableSet({self._con.userIri})
                    if not self._values.get(Xsd_QName('oldap:lastModificationDate', validate=False)):
                        self._values[Xsd_QName('oldap:lastModificationDate', validate=False)] = ObservableSet({timestamp})
                set_values(sc.properties)
                for iri, prop in sc.properties.items():
                    self.properties[iri] = prop

        if self.superclass:
            process_superclasses(self.superclass)

        set_values(self.properties)

        for propname in self.properties.keys():
            setattr(ResourceInstance, propname.fragment, property(
                partial(ResourceInstance.__get_value, attr=propname),
                partial(ResourceInstance.__set_value, attr=propname),
                partial(ResourceInstance.__del_value, attr=propname)))
        self.clear_changeset()

    def __setitem__(self, key: Xsd_QName, value: ValueType | Xsd | None):
        if self._values.get(key):
            self.__set_value(value, key)
        else:
            raise OldapErrorValue(f'{self.name}: Property {key} does not exist.')

    def __getitem__(self, key: Xsd_QName) -> ValueType | Xsd | None:
        if self._values.get(key):
            return self.__get_value(key)
        else:
            raise OldapErrorValue(f'{self.name}: Property {key} does not exist.')

    def __delitem__(self, key: Xsd_QName):
        if self._values.get(key):
            self.__del_value(key)

    def get(self, key: Xsd_QName) -> ValueType | Xsd | None:
        if self._values.get(key):
            return self._values[key]
        else:
            return None

    def validate_value(self, values: ValueType, property: PropertyClass):
        """
        Validates a set of values against a given property and its constraints. This function ensures
        that the input values conform to specific constraints defined by the property. The function
        verifies conformance to constraints such as language requirements, predefined inclusions,
        length limits, pattern matching, exclusivity, inclusivity, and relational comparisons
        (less than or less than or equals).

        :param values: A set of values to be validated.
        :param property: The property against which the values are being validated, containing
            the constraints that the values must meet.
        :return: None if the values successfully pass all validations.
        :raises OldapErrorInconsistency: Raised when the values do not conform to the defined property
            constraints due to inconsistencies or invalid data types.
        :raises OldapErrorValue: Raised when the values explicitly violate specific constraints
            such as inclusion constraints or invalid language values.
        """
        if property.get(PropClassAttr.LANGUAGE_IN):  # testing for LANGUAGE_IN conformance
            if not isinstance(values, LangString):
                raise OldapErrorInconsistency(f'Property {property.property_class_iri} with LANGUAGE_IN requires datatype rdf:langstring, got {type(values).__name__}.')
            for lang, dummy in values.items():
                if not lang in property[PropClassAttr.LANGUAGE_IN]:
                    raise OldapErrorValue(f'Property {property.property_class_iri} with LANGUAGE_IN={property[PropClassAttr.LANGUAGE_IN]} has invalid language "{lang.value}"')
        if property.get(PropClassAttr.UNIQUE_LANG):
            return  # TODO: LangString does not yet allow multiple entries of the same language...
        if property.get(PropClassAttr.IN):
            for val in values:
                if not val in property[PropClassAttr.IN]:
                    raise OldapErrorValue(f'Property {property} with IN={property[PropClassAttr.IN]} has invalid value "{val}"')
        if property.get(PropClassAttr.MIN_LENGTH):
            for val in values:
                l = 0
                try:
                    l = len(val)
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_LENGTH={property[PropClassAttr.MIN_LENGTH]} has no length.')
                if l < property[PropClassAttr.MIN_LENGTH]:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_LENGTH={property[PropClassAttr.MIN_LENGTH]} has length "{len(val)}".')
        if property.get(PropClassAttr.MAX_LENGTH):
            for val in values:
                l = 0
                try:
                    l = len(val)
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MAX_LENGTH={property[PropClassAttr.MAX_LENGTH]} has no length.')
                if l > property[PropClassAttr.MAX_LENGTH]:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_LENGTH={property[PropClassAttr.MAX_LENGTH]} has length "{len(val)}".')
        if property.get(PropClassAttr.PATTERN):
            for val in values:
                if not re.fullmatch(str(property[PropClassAttr.PATTERN]), str(val)):
                    raise OldapErrorInconsistency(f'Property {property} with PATTERN={property[PropClassAttr.PATTERN]} does not conform ({val}).')
        if property.get(PropClassAttr.MIN_EXCLUSIVE):
            for val in values:
                v: bool | None = None
                try:
                    v = val > property[PropClassAttr.MIN_EXCLUSIVE]
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_EXCLUSIVE]} cannot be compared to "{val}".')
                if not v:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_EXCLUSIVE]} has invalid "{val}".')
        if property.get(PropClassAttr.MIN_INCLUSIVE):
            for val in values:
                v: bool | None = None
                try:
                    v = val >= property[PropClassAttr.MIN_INCLUSIVE]
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_INCLUSIVE]} cannot be compared to "{val}".')
                if not v:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_INCLUSIVE]} has invalid "{val}".')
        if property.get(PropClassAttr.MAX_EXCLUSIVE):
            for val in values:
                v: bool | None = None
                try:
                    v = val < property[PropClassAttr.MAX_EXCLUSIVE]
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MAX_EXCLUSIVE={property[PropClassAttr.MAX_EXCLUSIVE]} cannot be compared to "{val}".')
                if not v:
                    raise OldapErrorInconsistency(f'Property {property} with MAX_EXCLUSIVE={property[PropClassAttr.MAX_EXCLUSIVE]} has invalid "{val}".')
        if property.get(PropClassAttr.MAX_INCLUSIVE):
            for val in values:
                v: bool | None = None
                try:
                    v = val <= property[PropClassAttr.MAX_INCLUSIVE]
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MAX_INCLUSIVE={property[PropClassAttr.MAX_INCLUSIVE]} cannot be compared to "{val}".')
                if not v:
                    raise OldapErrorInconsistency(f'Property {property} with MAX_INCLUSIVE={property[PropClassAttr.MAX_INCLUSIVE]} has invalid "{val}".')
        if property.get(PropClassAttr.LESS_THAN):
            other_values = self._values.get(property[PropClassAttr.LESS_THAN])
            if other_values is not None:
                b: bool | None = None
                try:
                    min_other_value = min(other_values)
                    max_value = max(values)
                    b = max_value < min_other_value
                except TypeError:
                    raise OldapErrorInconsistency(
                        f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} cannot be compared to "{values}".')
                if not b:
                    raise OldapErrorInconsistency(
                        f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} has invalid value: "{max_value}" NOT LESS_THAN "{min_other_value}".')
        if property.get(PropClassAttr.LESS_THAN_OR_EQUALS):
            other_values = self._values.get(property[PropClassAttr.LESS_THAN])
            if other_values is not None:
                b: bool | None = None
                try:
                    min_other_value = min(other_values)
                    max_value = max(values)
                    b = max_value <= min_other_value
                except TypeError:
                    raise OldapErrorInconsistency(
                        f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN_OR_EQUALS]} cannot be compared "{values}".')
                if not b:
                    raise OldapErrorInconsistency(
                        f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN_OR_EQUALS]} has invalid value: "{max_value}" NOT LESS_THAN "{min_other_value}".')

    def notifier(self, prop_iri: Xsd_QName):
        hasprop = self.properties[prop_iri]
        self.validate_value(self._values[prop_iri], hasprop)

        if hasprop.get(HasPropertyAttr.MIN_COUNT):  # testing for MIN_COUNT conformance
            n = len(self._values[prop_iri])
            if n < hasprop[HasPropertyAttr.MIN_COUNT]:
                self._values[prop_iri].undo()
                raise OldapErrorValue(
                    f'{self.name}: Property {prop_iri} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} has not enough values (n={n}).')

        if hasprop.get(HasPropertyAttr.MAX_COUNT):  # testing for MAX_COUNT conformance
            n = len(self._values[prop_iri])
            if n > hasprop[HasPropertyAttr.MAX_COUNT]:
                self._values[prop_iri].undo()
                raise OldapErrorValue(
                    f'{self.name}: Property {prop_iri} with MAX_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} has to many values (n={n}).')

        self._changeset[prop_iri] = AttributeChange(None, Action.MODIFY)

    def check_for_permissions(self, permission: AdminPermission) -> tuple[bool, str]:
        #
        # First we check if the logged-in user ("actor") has the permission to create a user for
        # the given project!
        #
        actor = self._con.userdata
        sysperms = actor.inProject.get(Iri('oldap:SystemProject', validate=False))
        if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
            #
            # user has root privileges!
            #
            return True, "OK – IS ROOT"
        else:
            perms = actor.inProject.get(self.project.projectIri)
            if permission in perms:
                return True, "OK",
            else:
                return False, f'Actor does not have {permission} in project "{self.project.projectShortName}".'

    def __get_value(self: Self, attr: Xsd_QName | str) -> Xsd | ValueType | None:
        #attr = Xsd_QName(prefix, fragment, validate=False)
        if not isinstance(attr, Xsd_QName):
            attr = Xsd_QName(attr)
        tmp = self._values.get(attr, None)
        if tmp is not None and str(attr) in {'oldap:createdBy', 'oldap:creationDate', 'oldap:lastModifiedBy', 'oldap:lastModificationDate'}:
            return next(iter(tmp))
        return tmp

    def __set_value(self: Self, value: ValueType | Xsd | None, attr: Xsd_QName | str) -> None:
        if not isinstance(attr, Xsd_QName):
            attr = Xsd_QName(attr)
        hasprop = self.properties.get(attr)

        #
        # Validate
        #
        if hasprop.get(HasPropertyAttr.MIN_COUNT):  # testing for MIN_COUNT conformance
            if hasprop[HasPropertyAttr.MIN_COUNT] > 0 and not value:
                raise OldapErrorValue(
                    f'{self.name}: Property {attr} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} is missing')
            elif isinstance(value, (list, tuple, set, ObservableSet)) and len(value) < hasprop[HasPropertyAttr.MIN_COUNT]:
                raise OldapErrorValue(
                    f'{self.name}: Property {attr} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} has not enough values')
        if hasprop.get(HasPropertyAttr.MAX_COUNT):  # testing for MAX_COUNT conformance
            if isinstance(value, (list, tuple, set, ObservableSet)) and len(value) > hasprop[HasPropertyAttr.MAX_COUNT]:
                raise OldapErrorValue(
                    f'{self.name}: Property {attr} with MAX_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} has to many values (n={len(value)})')
        if value:
            if isinstance(value, LangString):
                self.validate_value(value, hasprop.prop)
            else:
                if isinstance(value, (list, tuple, set, ObservableSet)):
                    for val in self._values[attr]:
                        self.validate_value(val, hasprop.prop)
                else:
                    self.validate_value(value, hasprop.prop)

        if not value:
            self._changeset[attr] = AttributeChange(self._values.get(attr), Action.DELETE)
            del self._values[attr]
        elif isinstance(value, (list, tuple, set, LangString)):  # we may have multiple values...
            if self._values.get(attr):
                self._changeset[attr] = AttributeChange(self._values.get(attr), Action.REPLACE)
            else:
                self._changeset[attr] = AttributeChange(None, Action.CREATE)
            if hasprop.prop.datatype == XsdDatatypes.langString:
                self._values[attr] = LangString(value, notifier=self.notifier, notify_data=attr)
            else:
                self._values[attr] = ObservableSet({
                    convert2datatype(x, hasprop.prop.datatype) for x in value
                }, notifier=self.notifier, notify_data=attr)
        else:
            if self._values.get(attr):
                self._changeset[attr] = AttributeChange(self._values.get(attr), Action.REPLACE)
            else:
                self._changeset[attr] = AttributeChange(None, Action.CREATE)
            try:
                self._values[attr] = ObservableSet({convert2datatype(value, hasprop.prop.datatype)})
            except TypeError:
                self._values[attr] = convert2datatype(value, hasprop.prop.datatype)

    def __del_value(self: Self, attr: Xsd_QName | str) -> None:
        if not isinstance(attr, Xsd_QName):
            attr = Xsd_QName(attr)
        hasprop = self.properties.get(attr)

        if hasprop.get(HasPropertyAttr.MIN_COUNT):  # testing for MIN_COUNT conformance
            if hasprop[HasPropertyAttr.MIN_COUNT] > 0:
                raise OldapErrorValue(f'{self.name}: Property {attr} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} cannot be deleted.')

        self._changeset[attr] = AttributeChange(self._values.get(attr), Action.DELETE)
        del self._values[attr]

    @property
    def iri(self) -> Iri:
        return self._iri

    @property
    def changeset(self) -> dict[Xsd_QName, AttributeChange]:
        return self._changeset

    def clear_changeset(self) -> None:
        for item in self._values:
            if hasattr(self._values[item], 'clear_changeset'):
                self._values[item].clear_changeset()
        self._changeset = {}


    def get_data_permission(self, context: Context, permission: DataPermission) -> bool:
        permission_query = context.sparql_context
        permission_query += f'''
        ASK {{
            GRAPH {self._graph}:data {{
                {self._iri.toRdf} oldap:grantsPermission ?permset .
            }}
            GRAPH oldap:admin {{
                {self._con.userIri.toRdf} oldap:hasPermissions ?permset .
                ?permset oldap:givesPermission ?DataPermission .
                ?DataPermission oldap:permissionValue ?permval .
            }}
            FILTER(?permval >= {permission.numeric.toRdf})
        }}'''

        if self._con.in_transaction():
            result = self._con.transaction_query(permission_query)
        else:
            result = self._con.query(permission_query)
        return result['boolean']

    def create(self, indent: int = 0, indent_inc: int = 4) -> str:
        """
        Generates an RDF/SPARQL INSERT DATA query for creating a resource with associated
        metadata and verifies permissions before execution.

        This method first checks for the necessary permissions to create a resource.
        It assigns metadata such as the creation and modification information if
        the resource name matches specific criteria. It ensures that no resource
        with the same IRI exists before proceeding with the resource creation.
        If a conflict is detected, an error is raised. The method finalizes the
        creation of the resource by executing a SPARQL query in a transactional context.

        :param indent: The current level of indentation (default is 0).
        :type indent: int
        :param indent_inc: The incremental spaces per indentation level (default is 4).
        :type indent_inc: int
        :return: A string representing the RDF/SPARQL query for creation.
        :rtype: str
        :raises OldapErrorNoPermission: If the user does not have the required permissions to create the resource.
        :raises OldapErrorAlreadyExists: If a resource with the same IRI already exists.
        :raises OldapError: For errors encountered during the SPARQL transaction execution.
        """
        result, message = self.check_for_permissions(AdminPermission.ADMIN_CREATE)
        if not result:
            raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTimeStamp()
        if self.name in ("Thing", "oldap:Thing"):
            self._values[Xsd_QName('oldap:createdBy', validate=False)] = ObservableSet({self._con.userIri})
            self._values[Xsd_QName('oldap:creationDate', validate=False)] = ObservableSet({timestamp})
            self._values[Xsd_QName('oldap:lastModifiedBy', validate=False)] = ObservableSet({self._con.userIri})
            self._values[Xsd_QName('oldap:lastModificationDate', validate=False)] = ObservableSet({timestamp})

        indent: int = 0
        indent_inc: int = 4

        blank = ''
        context = Context(name=self._con.context_name)

        #
        # Test if a resource with the same IRI already exists!
        #
        sparql0 = context.sparql_context
        sparql0 += f'ASK {{ GRAPH {self._graph}:data {{ {self._iri.toRdf} ?p ?o }} }}'

        sparql = context.sparql_context
        sparql += f'{blank:{indent * indent_inc}}INSERT DATA {{'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}GRAPH {self._graph}:data {{'

        sparql += f'\n{blank:{(indent + 2) * indent_inc}}{self._iri.toRdf} a {self.name}'

        for prop_iri, values in self._values.items():
            if self.properties.get(prop_iri) and self.properties[prop_iri].prop.datatype == XsdDatatypes.QName:
                qnames = {f'"{x}"^^xsd:QName' for x in values}
                qnames_rdf = ', '.join(qnames)
                sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}{prop_iri.toRdf} {qnames_rdf}'
            else:
                sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}{prop_iri.toRdf} {values.toRdf}'
        sparql += f' .\n{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql += f'{blank:{indent * indent_inc}}}}\n'

        self._con.transaction_start()
        try:
            result = self._con.transaction_query(sparql0)
            if result['boolean']:
                self._con.transaction_abort()
                raise OldapErrorAlreadyExists(f'Resource with IRI {self._iri} already exists.')
            self._con.transaction_update(sparql)
        except OldapError as e:
            print(sparql)
            self._con.transaction_abort()
            raise
        self._con.transaction_commit()


    @classmethod
    def read(cls,
             con: IConnection,
             iri: Iri) -> Self:
        """
        Reads an object of the specific class type from the RDF data source using the provided connection
        and IRI. Validates that the retrieved RDF data matches the expected class type and ensures proper
        permissions have been granted to access the resource.

        :param con: The connection object to the data source (IConnection).
        :param iri: The IRI of the resource to read (Iri).
        :return: An object of the invoking class loaded with data from the RDF resource.

        :raises OldapErrorInconsistency: If an expected QName is not found or if the retrieved object type
                                          does not match the expected class type.
        :raises OldapErrorNotFound: If the resource associated with the provided IRI is not found.
        """
        graph = cls.project.projectShortName
        context = Context(name=con.context_name)
        sparql = context.sparql_context
        sparql += f'''
        SELECT ?predicate ?value
        WHERE {{
            GRAPH {graph}:data {{
                {iri.toRdf} ?predicate ?value .
                {iri.toRdf} oldap:grantsPermission ?permset .
            }}
            GRAPH oldap:admin {{
                {con.userIri.toRdf} oldap:hasPermissions ?permset .
                ?permset oldap:givesPermission ?DataPermission .
                ?DataPermission oldap:permissionValue ?permval .
            }}
            FILTER(?permval >= {DataPermission.DATA_VIEW.numeric.toRdf})
        }}'''
        jsonres = con.query(sparql)
        res = QueryProcessor(context, jsonres)
        objtype = None
        kwargs: dict[str, Any] = {}
        for r in res:
            if r['predicate'] == 'rdf:type':
                if r['value'].is_qname:
                    objtype = r['value'].as_qname.fragment
                else:
                    raise OldapErrorInconsistency(f"Expected QName as value, got {r['value']}")
            else:
                if r['predicate'].is_qname:
                    if kwargs.get(r['predicate'].as_qname.fragment):
                        if isinstance(kwargs[r['predicate'].as_qname.fragment], set):
                            kwargs[r['predicate'].as_qname.fragment].add(r['value'])
                        else:
                            kwargs[r['predicate'].as_qname.fragment] = r['value']
                    else:
                        try:
                            kwargs[r['predicate'].as_qname.fragment] = {r['value']}
                        except TypeError:
                            kwargs[r['predicate'].as_qname.fragment] = r['value']
                else:
                    raise OldapErrorInconsistency(f"Expected QName as predicate, got {r['predicate']}")
        if objtype is None:
            raise OldapErrorNotFound(f'Resource with iri <{iri}> not found.')
        if cls.__name__ != objtype:
            raise OldapErrorInconsistency(f'Expected class {cls.__name__}, got {objtype} instead.')
        return cls(iri=iri, **kwargs)

    def update(self, indent: int = 0, indent_inc: int = 4) -> None:
        """
        Updates the current resource by creating and executing SPARQL queries to modify, insert, or
        delete data based on the changeset. This method also verifies user permissions and ensures
        data consistency using timestamps. It commits modifications within a transaction while
        rolling back changes upon encountering errors.

        The `update` method processes changes in resource attributes specified in a changeset. It
        constructs appropriate SPARQL queries for CREATE, MODIFY, and DELETE actions for each
        field. Additionally, it updates the last modification timestamp for the resource and records
        the user performing the update. If permissions are insufficient or an update fails, the
        transaction is aborted.

        :param indent: The initial level of indentation for the generated SPARQL queries.
        :type indent: int
        :param indent_inc: The incremental level of indentation for nested sections in SPARQL
                           queries.
        :type indent_inc: int
        :return: None
        :rtype: None
        """
        admin_resources, message = self.check_for_permissions(AdminPermission.ADMIN_RESOURCES)

        context = Context(name=self._con.context_name)
        blank = ''
        timestamp = Xsd_dateTimeStamp()

        sparql_list = []
        required_permission = DataPermission.DATA_EXTEND
        for field, change in self._changeset.items():
            if field == 'oldap:grantsPermission':
                required_permission = DataPermission.DATA_PERMISSIONS
            if change.action == Action.MODIFY:
                continue  # will be processed below!
            if change.action != Action.CREATE:
                if required_permission < DataPermission.DATA_UPDATE:
                    required_permission = DataPermission.DATA_UPDATE
            sparql = f'# Processing field "{field}"\n'
            sparql += f'{blank:{indent * indent_inc}}WITH {self._graph}:data\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                if isinstance(change.old_value, (ObservableSet, LangString)):
                    for val in change.old_value:
                        sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {val.toRdf} .\n'
                else:
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                if isinstance(change.old_value, (ObservableSet, LangString)):
                    for val in self._values[field]:
                        sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {val.toRdf} .\n'
                else:
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {self._values[field].toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self._iri.toRdf} as ?res_iri)\n'
            if change.action != Action.CREATE:
                if isinstance(change.old_value, (ObservableSet, LangString)):
                    for val in change.old_value:
                        sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {val.toRdf} .\n'
                else:
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {change.old_value.toRdf} .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)

        for field, change in self._changeset.items():
            if change.action != Action.MODIFY:
                continue  # has been processed above
            if self.properties[field].prop.datatype == XsdDatatypes.langString:
                sparqls = self._values[field].update(graph=Xsd_QName(self._graph, 'data'),
                                                     subject=self._iri,
                                                     field=field.as_qname)
                for lang, lchange in self._values[field].changeset.items():
                    if lchange.action != Action.CREATE:
                        if required_permission < DataPermission.DATA_UPDATE:
                            required_permission = DataPermission.DATA_UPDATE
                sparql_list.extend(sparqls)
            else:
                #
                # first we rectify the datatype of all "new" values added to the set
                #
                newset = {convert2datatype(x, self.properties[field].prop.datatype) for x in self._values[field]}
                self._values[field] = ObservableSet(newset, old_value=self._values[field].old_value, notifier=self.notifier, notify_data=field)
                sparqls = self._values[field].update_sparql(graph=Iri(f'{self._graph}:data'),
                                                            subject=self._iri,
                                                            field=field)
                sparql_list.extend(sparqls)

        sparql = context.sparql_context
        sparql += f'# Updating resource "{self._iri}"\n'
        sparql += " ;\n".join(sparql_list)

        modtime_update = context.sparql_context
        modtime_update += f'''
        WITH {self._graph}:data
        DELETE {{
            ?res oldap:lastModificationDate {self.lastModificationDate.toRdf} .
            ?res oldap:lastModifiedBy ?contributor .
        }}
        INSERT {{
            ?res oldap:lastModificationDate {timestamp.toRdf} .
            ?res oldap:lastModifiedBy {self._con.userIri.toRdf} .
        }}
        WHERE {{
            BIND({self._iri.toRdf} as ?res)
            ?res oldap:lastModificationDate {self.lastModificationDate.toRdf} .
            ?res oldap:lastModifiedBy ?contributor .
        }}
        '''

        context = Context(name=self._con.context_name)
        modtime_get = context.sparql_context
        modtime_get += f"""
        SELECT ?modified
        FROM {self._graph}:data
        WHERE {{
            {self._iri.toRdf} oldap:lastModificationDate ?modified
        }}
        """

        self._con.transaction_start()
        #
        # Test permission for Action.REPLACE
        #
        if not admin_resources:
            if not self.get_data_permission(context, required_permission):
                self._con.transaction_abort()
                raise OldapErrorNoPermission(f'No permission to update resource "{self._iri}"')
        try:
            self._con.transaction_update(sparql)
            self._con.transaction_update(modtime_update)
            jsonobj = self._con.transaction_query(modtime_get)
            res = QueryProcessor(context, jsonobj)
            modtime = res[0]['modified']
            if timestamp != modtime:
                raise OldapErrorUpdateFailed(f"Update failed! Timestamp does not match (modtime={modtime}, timestamp={timestamp}).")
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self.clear_changeset()

    def delete(self) -> None:
        """
        Deletes the specified resource represented by the object's IRI from the associated graph
        if the proper permissions are granted and the resource is not in use. This method initiates
        a transactional sequence to ensure data consistency and rollback in case of an error.
        The process includes checking permissions, querying whether the resource is in use,
        and finally issuing a SPARQL DELETE query to remove the resource.

        Raises appropriate exceptions if the resource is in use, permissions are insufficient,
        or any other transaction-related error occurs.

        :raises OldapErrorNoPermission: If the user lacks the necessary permissions to
            delete the resource.
        :raises OldapErrorInUse: If the resource is currently in use and cannot be deleted.
        :raises OldapError: If any error occurs during transaction start, query, update,
            or commit phases.
        """
        admin_resources, message = self.check_for_permissions(AdminPermission.ADMIN_RESOURCES)

        context = Context(name=self._con.context_name)
        inuse = context.sparql_context
        inuse += f"""
        ASK {{
            GRAPH ?g {{
                ?res ?prop {self._iri.toRdf} .
            }}
        }}
        """

        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        DELETE WHERE {{
            GRAPH {self._graph}:data {{
                {self._iri.toRdf} ?prop ?val .
            }}
        }} 
        """

        self._con.transaction_start()
        if not admin_resources:
            if not self.get_data_permission(context, DataPermission.DATA_DELETE):
                self._con.transaction_abort()
                raise OldapErrorNoPermission(f'No permission to update resource "{self._iri}"')
        try:
            result = self._con.transaction_query(inuse)
            if result['boolean']:
                raise OldapErrorInUse(f'Resource "{self._iri}" is in use and cannot be deleted.')
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_update(sparql)
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise

    @staticmethod
    def read_data(con: IConnection, projectShortName: Xsd_NCName | str, iri: Iri | str) -> dict[Xsd_QName, Any]:
        """
        Reads and processes data from a given IConnection instance based on a specified
        project short name and a resource IRI. This method fetches the data from a graph
        in the RDF database, verifies user permissions, and returns selected data in a
        dictionary format.

        :param con: The connection object used to query the RDF data.
        :type con: IConnection
        :param projectShortName: The short name of the project used to identify the graph
            in the RDF store. Accepts either Xsd_NCName or a string.
        :param iri: The IRI of the resource being queried. Accepts either Iri or a string.
        :return: A dictionary mapping predicates to their associated values from the
            selected resource.
        :rtype: dict[Xsd_QName, Any]
        :raises OldapErrorInconsistency: If a non-QName predicate is found in the results.
        :raises OldapErrorNotFound: If the resource with the given IRI is not found.
        """
        if not isinstance(iri, Iri):
            iri = Iri(iri, validate=True)
        if not isinstance(projectShortName, Xsd_NCName):
            graph = Xsd_NCName(projectShortName)
        else:
            graph = projectShortName

        context = Context(name=con.context_name)
        sparql = context.sparql_context
        sparql += f'''
        SELECT ?predicate ?value
        WHERE {{
            GRAPH {graph}:data {{
                {iri.toRdf} ?predicate ?value .
                {iri.toRdf} oldap:grantsPermission ?permset .
            }}
            GRAPH oldap:admin {{
            	{con.userIri.toRdf} oldap:hasPermissions ?permset .
            	?permset oldap:givesPermission ?DataPermission .
            	?DataPermission oldap:permissionValue ?permval .
            }}
            FILTER(?permval >= {DataPermission.DATA_VIEW.numeric.toRdf})
        }}
        '''
        jsonres = con.query(sparql)
        res = QueryProcessor(context, jsonres)
        data = {}
        for r in res:
            if r['predicate'].is_qname:
                if not data.get(r['predicate'].as_qname):
                    data[r['predicate'].as_qname] = []
                data[str(r['predicate'].as_qname)].append(str(r['value']))
            else:
                raise OldapErrorInconsistency(f"Expected QName as predicate, got {r['predicate']}")
        if not data.get('rdf:type'):
            raise OldapErrorNotFound(f'Resource with iri <{iri}> not found.')
        return data

    @staticmethod
    def search_fulltext(con: IConnection,
                        projectShortName: Xsd_NCName | str,
                        s: str,
                        resClass: Xsd_QName | str | None = None,
                        countOnly: bool = False,
                        sortBy: SortBy | None = None,
                        limit: int = 100,
                        offset: int = 0,
                        indent: int = 0, indent_inc: int = 4) -> int | dict[Iri, dict[str, Xsd]]:
        """
        Search for entities in a project using a full-text search query, with optional filters and sorting
        based on resource class and properties.

        :param con: The connection instance to interact with the SPARQL endpoint or database.
        :type con: IConnection
        :param projectShortName: The project short name (identifier). Can be provided as an Xsd_NCName or a
            string. If a string, it is automatically validated and converted to an Xsd_NCName.
        :type projectShortName: Xsd_NCName | str
        :param s: The full-text search query to identify entities matching the specified string values.
        :type s: str
        :param resClass: An optional entity class (Xsd_QName) used to filter results. If provided as a string,
            it is validated and converted to Xsd_QName.
        :type resClass: Xsd_QName | str | None
        :param countOnly: If True, only the count of matching entities will be returned. Defaults to False.
        :type countOnly: bool
        :param sortBy: Optional sorting criterion for the results, such as creation date, last modification date,
            or property values. Defaults to None.
        :type sortBy: SortBy | None
        :param limit: The maximum number of records to return when count_only is False. Defaults to 100.
        :type limit: int
        :param offset: The number of records to skip for paginated queries. Defaults to 0.
        :type offset: int
        :param indent: Initial indentation level for the generated SPARQL query. Defaults to 0.
        :type indent: int
        :param indent_inc: Incremental indentation used during SPARQL query generation. Defaults to 4.
        :type indent_inc: int
        :return: If count_only is True, an integer representing the count of matching entities. If count_only is
            False, a dictionary with entity IRIs as keys and their respective property-to-value mappings
            as values.
        :rtype: int | dict[Iri, dict[str, Xsd]]

        The SPARQL generated is as follows

        ```sparql
        SELECT DISTINCT ?s ?t ?p ?o ?creationDate
        WHERE {
            GRAPH test:data {
                ?s oldap:creationDate ?creationDate .
                ?s ?p ?o .
                ?s rdf:type ?t .
                ?s rdf:type test:Book .
                ?s oldap:grantsPermission ?permset .
            }
            FILTER(isLiteral(?o) && (datatype(?o) = xsd:string || datatype(?o) = rdf:langString || lang(?o) != ""))
            FILTER(CONTAINS(LCASE(STR(?o)), "gaga"))  # case-insensitive substring match
            GRAPH oldap:admin {
                <https://orcid.org/0000-0003-1681-4036> oldap:hasPermissions ?permset .
                ?permset oldap:givesPermission ?DataPermission .
                ?DataPermission oldap:permissionValue ?permval .
            }
            FILTER(?permval >= "2"^^xsd:integer)
        }
        ORDER BY ASC(?creationDate)LIMIT 100 OFFSET 0
        ```
        """
        if not isinstance(projectShortName, Xsd_NCName):
            graph = Xsd_NCName(projectShortName, validate=True)
        else:
            graph = projectShortName
        if resClass and not isinstance(resClass, Xsd_QName):
            resClass = Xsd_QName(resClass, validate=True)

        blank = ''
        context = Context(name=con.context_name)
        sparql = context.sparql_context

        if (countOnly):
            sparql += f'{blank:{indent * indent_inc}}SELECT (COUNT(DISTINCT ?s) as ?numResult)'
        else:
            sparql += f'{blank:{indent * indent_inc}}SELECT DISTINCT ?s ?t ?p ?o'
            if sortBy:
                if sortBy == SortBy.CREATED:
                    sparql += ' ?creationDate'
                if sortBy == SortBy.LASTMOD:
                    sparql += '?lastModificationDate'
        sparql += f'\n{blank:{indent * indent_inc}}WHERE {{'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}GRAPH {graph}:data {{'
        if sortBy:
            if sortBy == SortBy.CREATED:
                sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s oldap:creationDate ?creationDate .'
            if sortBy == SortBy.LASTMOD:
                sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s oldap:lastModificationDate ?lastModificationDate .'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s ?p ?o .'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s rdf:type ?t .'
        if resClass:
            sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s rdf:type {resClass} .'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s oldap:grantsPermission ?permset .'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}}}'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}FILTER(isLiteral(?o) && (datatype(?o) = xsd:string || datatype(?o) = rdf:langString || lang(?o) != ""))'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}FILTER(CONTAINS(LCASE(STR(?o)), "{s}"))  # case-insensitive substring match'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}{con.userIri.toRdf} oldap:hasPermissions ?permset .'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}?permset oldap:givesPermission ?DataPermission .'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}?DataPermission oldap:permissionValue ?permval .'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}}}'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}FILTER(?permval >= {DataPermission.DATA_VIEW.numeric.toRdf})'
        sparql += f'\n{blank:{indent * indent_inc}}}}'

        if sortBy:
            if sortBy == SortBy.CREATED:
                sparql += f'\n{blank:{indent * indent_inc}}ORDER BY ASC(?creationDate)'
            if sortBy == SortBy.LASTMOD:
                sparql += f'\n{blank:{indent * indent_inc}}ORDER BY ASC(?lastModificationDate)'
            if sortBy == SortBy.PROPVAL:
                sparql += f'\n{blank:{indent * indent_inc}}ORDER BY ASC(LCASE(STR(?o)))'

        if not countOnly:
            sparql += f'\n{blank:{indent * indent_inc}}LIMIT {limit} OFFSET {offset}'
        sparql += '\n'


        try:
            jsonres = con.query(sparql)
        except OldapError:
            print(sparql)
            raise
        res = QueryProcessor(context, jsonres)
        if countOnly:
            return res[0]['numResult']
        else:
            result = {}
            for r in res:
                tmp = {
                    r['p']: r['o'],
                    Xsd_QName('owl:Class'): r['t']
                }
                if sortBy:
                    if sortBy == SortBy.CREATED:
                        tmp['oldap:creationDate'] = r['creationDate'],
                    if sortBy == SortBy.LASTMOD:
                        tmp['oldap:lastModificationDate'] = r['lastModificationDate']
                result[r['s']] = tmp
            return result

    @staticmethod
    def all_resources(con: IConnection,
                      projectShortName: Xsd_NCName | str,
                      resClass: Xsd_QName | str,
                      includeProperties: list[Xsd_QName] | None = None,
                      count_only: bool = False,
                      sortBy: SortBy | None = None,
                      limit: int = 100,
                      offset: int = 0,
                      indent: int = 0, indent_inc: int = 4) -> dict[Iri, dict[str, Xsd]]:
        """
        Retrieves all resources matching the specified parameters from a data store using a SPARQL query.
        Depending on the `count_only` flag, it can return either a count of matching resources or detailed
        information about each resource.

        :param con: The connection object used to execute the SPARQL query.
        :type con: IConnection
        :param projectShortName: The short name of the project being queried.
        :type projectShortName: Xsd_NCName | str
        :param resClass: The resource class to filter the results by.
        :type resClass: Xsd_QName | str
        :param includeProperties: A list of resource properties to include in the query and results.
        :type includeProperties: list[Xsd_QName] | None
        :param count_only: If True, returns only the count of matching resources. Defaults to False.
        :type count_only: bool
        :param sortBy: The sorting criterion for the results. Defaults to None.
        :type sortBy: SortBy | None
        :param limit: The maximum number of results to retrieve. Defaults to 100.
        :type limit: int
        :param offset: The starting point for result retrieval in pagination. Defaults to 0.
        :type offset: int
        :param indent: The initial level of indentation for the SPARQL query strings. Defaults to 0.
        :type indent: int
        :param indent_inc: The incremental level of indentation for the SPARQL query strings. Defaults to 4.
        :type indent_inc: int
        :return: A dictionary where keys are resource IRIs and values are dictionaries containing resource
            properties and their corresponding values. If `count_only` is True, returns the total count of
            matching resources as an integer.
        :rtype: dict[Iri, dict[str, Xsd]]
        :raises OldapError: If there is an error during query execution in the connection object.
        """
        if not isinstance(projectShortName, Xsd_NCName):
            graph = Xsd_NCName(projectShortName, validate=True)
        else:
            graph = projectShortName
        if not isinstance(resClass, Xsd_QName):
            resClass = Xsd_QName(resClass, validate=True)

        blank = ''
        context = Context(name=con.context_name)
        sparql = context.sparql_context

        if (count_only):
            sparql += f'{blank:{indent * indent_inc}}SELECT (COUNT(DISTINCT ?s) as ?numResult)'
        else:
            sparql += f'{blank:{indent * indent_inc}}SELECT DISTINCT ?s'
            if includeProperties:
                for index, item in enumerate(includeProperties):
                    sparql += f' ?o{index}'
            if sortBy:
                if sortBy == SortBy.CREATED:
                    sparql += ' ?creationDate'
                if sortBy == SortBy.LASTMOD:
                    sparql += '?lastModificationDate'
        sparql += f'\n{blank:{indent * indent_inc}}WHERE {{'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}GRAPH {graph}:data {{'
        if sortBy:
            if sortBy == SortBy.CREATED:
                sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s oldap:creationDate ?creationDate .'
            if sortBy == SortBy.LASTMOD:
                sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s oldap:lastModificationDate ?lastModificationDate .'
        if includeProperties:
            for index, prop in enumerate(includeProperties):
                sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s {prop} ?o{index} .'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s rdf:type {resClass} .'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}?s oldap:grantsPermission ?permset .'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}}}'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}{con.userIri.toRdf} oldap:hasPermissions ?permset .'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}?permset oldap:givesPermission ?DataPermission .'
        sparql += f'\n{blank:{(indent + 2) * indent_inc}}?DataPermission oldap:permissionValue ?permval .'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}}}'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}FILTER(?permval >= {DataPermission.DATA_VIEW.numeric.toRdf})'
        sparql += f'\n{blank:{indent * indent_inc}}}}'

        if sortBy:
            if sortBy == SortBy.CREATED:
                sparql += f'\n{blank:{indent * indent_inc}}ORDER BY ASC(?creationDate)'
            if sortBy == SortBy.LASTMOD:
                sparql += f'\n{blank:{indent * indent_inc}}ORDER BY ASC(?lastModificationDate)'
            if sortBy == SortBy.PROPVAL:
                sparql += f'\n{blank:{indent * indent_inc}}ORDER BY'
                if includeProperties:
                    for index, item in enumerate(includeProperties):
                        sparql += f'\n{blank:{indent * indent_inc}} ?o{index}'

        if not count_only:
            sparql += f'\n{blank:{indent * indent_inc}}LIMIT {limit} OFFSET {offset}'
        sparql += '\n'

        try:
            jsonres = con.query(sparql)
        except OldapError:
            print(sparql)
            raise
        res = QueryProcessor(context, jsonres)
        if count_only:
            return res[0]['numResult']
        else:
            result = {}
            for r in res:
                tmp = {}
                if includeProperties:
                    for index, item in enumerate(includeProperties):
                        tmp[item] = r[f'o{index}']

                if sortBy:
                    if sortBy == SortBy.CREATED:
                        tmp[Xsd_QName('oldap:creationDate')] = r['creationDate']
                    if sortBy == SortBy.LASTMOD:
                        tmp[Xsd_QName('oldap:lastModificationDate')] = r['lastModificationDate']
                result[r['s']] = tmp
            return result


    def toJsonObject(self) -> dict[str, list[str] | str]:
        result = {'iri': str(self._iri)}
        for propiri, values in self._values.items():
            if str(propiri) in {'oldap:createdBy', 'oldap:creationDate', 'oldap:lastModificationDate', 'oldap:lastModifiedBy'}:
                iterator = iter(values)
                result[str(propiri)] = str(next(iterator))
                continue

            result[str(propiri)] = [str(v) for v in values]
        return result


class ResourceInstanceFactory:
    """
    Represents a factory for creating instances of resources in a specific project.

    The `ResourceInstanceFactory` class is used for creating resource instances
    associated with a data model and project. It provides an interface to dynamically
    generate classes for resource objects with properties and inheritance derived
    from the project's data model.

    :ivar _con: The connection interface used to interact with the backend.
    :type _con: IConnection
    :ivar _project: Represents the project associated with the factory.
    :type _project: Project
    :ivar _datamodel: Represents the data model associated with the project.
    :type _datamodel: DataModel
    """
    _con: IConnection
    _project: Project
    _sharedProject: Project
    _datamodel: DataModel
    _sharedModel: DataModel

    def __init__(self,
                 con: IConnection,
                 project: Project | Iri | Xsd_NCName | str):
        self._con = con
        if isinstance(project, Project):
            self._project = project
        else:
            self._project = Project.read(self._con, project)
        self._sharedProject = Project.read(self._con, "oldap:SharedProject")

        self._datamodel = DataModel.read(con=self._con, project=self._project)
        self._sharedModel = DataModel.read(con=self._con, project=self._sharedProject)

    def createObjectInstance(self, name: Xsd_NCName | Xsd_QName | str) -> Type:  ## ToDo: Get name automatically from IRI
        if isinstance(name, Xsd_QName):
            classiri = name
        else:
            try:
                classiri = Xsd_QName(name, validate=True)
            except (ValueError, OldapErrorValue):
                if not isinstance(name, Xsd_NCName):
                    name = Xsd_NCName(name, validate=True)
                classiri = Xsd_QName(self._project.projectShortName, name)
        resclass = self._datamodel.get(classiri)
        if not resclass:
            resclass = self._sharedModel.get(classiri)
        if resclass is None:
            raise OldapErrorNotFound(f'Given Resource Class "{classiri}" not found.')
        return type(str(name), (ResourceInstance,), {
            '_con': self._con,
            'project': self._project,
            'name': resclass.owl_class_iri,
            'factory': self,
            'properties': resclass.properties,
            'superclass': resclass.superclass})


    def read(self, iri: Iri | str) -> ResourceInstance:
        if not isinstance(iri, Iri):
            iri = Iri(iri, validate=True)
        graph = self._project.projectShortName
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f'''
        SELECT ?predicate ?value
        WHERE {{
        	BIND({iri.toRdf} as ?iri)
            GRAPH {graph}:data {{
                ?iri ?predicate ?value .
                ?iri oldap:grantsPermission ?permset .
            }}
            BIND({self._con.userIri.toRdf} as ?user)
            GRAPH oldap:admin {{
            	?user oldap:hasPermissions ?permset .
            	?permset oldap:givesPermission ?DataPermission .
            	?DataPermission oldap:permissionValue ?permval .
            }}
            FILTER(?permval >= {DataPermission.DATA_VIEW.numeric.toRdf})
        }}'''
        jsonres = self._con.query(sparql)
        res = QueryProcessor(context, jsonres)
        objtype = None
        kwargs: dict[str, Any] = {}
        for r in res:
            if r['predicate'] == 'rdf:type':
                if r['value'].is_qname:
                    objtype = r['value'].as_qname.fragment
                else:
                    raise OldapErrorInconsistency(f"Expected QName as value, got {r['value']}")
            else:
                if r['predicate'].is_qname:
                    if kwargs.get(r['predicate'].as_qname.fragment):
                        if isinstance(kwargs[r['predicate'].as_qname.fragment], set):
                            kwargs[r['predicate'].as_qname.fragment].add(r['value'])
                        else:
                            kwargs[r['predicate'].as_qname.fragment] = r['value']
                    else:
                        try:
                            kwargs[r['predicate'].as_qname.fragment] = {r['value']}
                        except TypeError:
                            kwargs[r['predicate'].as_qname.fragment] = r['value']
                else:
                    raise OldapErrorInconsistency(f"Expected QName as predicate, got {r['predicate']}")
        if objtype is None:
            raise OldapErrorNotFound(f'Resource with iri <{iri}> not found.')
        Instance = self.createObjectInstance(objtype)
        return Instance(iri=iri, **kwargs)








